"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const messages_1 = require("../../util/messages");
const dielAstTypes_1 = require("../../parser/dielAstTypes");
const DielAstGetters_1 = require("../DielAstGetters");
const syntaxSugar_1 = require("./syntaxSugar");
/**
 * Find all the top level selections for:
 * - views
 * - outputs
 *
 * Possibly also joins for
 * - crossfilters
 *
 * Do the copy pass for:
 * - inputs
 * - dynamicTables
 * @param ast diel ast
 */
function ApplyTemplates(ast) {
    // note: i think the concat should be fine with modifying in place?
    DielAstGetters_1.GetAllDerivedViews(ast).map(r => {
        TryToApplyTemplate(r.selection);
        // also apply syntax sugar..
        syntaxSugar_1.ApplyLatestToDerivedRelation(r);
    });
    // defined here since it needs to access the global definition
    // and the copy pass
    DielAstGetters_1.GetAllDielDefinedOriginalRelations(ast).map(r => TryToCopyRelationSpec(ast, r));
}
exports.ApplyTemplates = ApplyTemplates;
function TryToCopyRelationSpec(ast, r) {
    if (r.copyFrom) {
        // make sure it's not copying from itself
        if (r.copyFrom === r.rName) {
            messages_1.ReportDielUserError(`You cannot copy ${r.rName} from itself!`);
        }
        // find the relation
        const sourceRelation = DielAstGetters_1.GetAllDielDefinedOriginalRelations(ast).filter(r => r.rName === r.copyFrom);
        if (sourceRelation.length === 0) {
            messages_1.ReportDielUserError(`The relation definition you are trying to copy from, ${r.copyFrom}, does not exist`);
        }
        else {
            r.columns = sourceRelation[0].columns;
        }
    }
}
exports.TryToCopyRelationSpec = TryToCopyRelationSpec;
/**
 * modify in place
 * @param ast
 */
function TryToApplyTemplate(ast) {
    if (!ast.templateSpec) {
        // there is no template to apply here!
        // LogInternalError(`Template variables not specified`);
        return;
    }
    if (ast.astType === dielAstTypes_1.AstType.RelationSelection) {
        _visitSelection(ast);
    }
    else {
        _visitJoinAst(ast);
    }
    function _changeString(inStr) {
        if (!inStr) {
            return inStr;
        }
        if ((inStr[0] === "{") && (inStr[inStr.length - 1] === "}")) {
            const varName = inStr.slice(1, inStr.length - 1);
            const varValue = ast.templateSpec.get(varName);
            if (varValue)
                return varValue;
            messages_1.LogInternalError(`${varName} was not found`);
            return undefined;
        }
        else {
            return inStr;
        }
    }
    function _visitColumnSelection(c) {
        // only change if it's a column
        // in theory this also needs to be recursive... but only deal with shallow stuff for now
        _visitExprAst(c.expr);
    }
    function _visitOrderByAst(c) {
        _visitExprAst(c.selection);
    }
    function _visitExprAst(e) {
        if (e) {
            if (e.exprType === dielAstTypes_1.ExprType.Column) {
                const c = e;
                const newCName = _changeString(c.columnName);
                if (newCName)
                    c.columnName = newCName;
                const newRName = _changeString(c.relationName);
                if (newRName)
                    c.relationName = newRName;
            }
            else if (e.exprType === dielAstTypes_1.ExprType.Func) {
                const f = e;
                // recursive!
                f.args.map(a => _visitExprAst(a));
            }
            else if (e.exprType === dielAstTypes_1.ExprType.Relation) {
                const r = e;
                // recursive!!
                _visitSelection(r.selection);
            }
        }
        else {
            messages_1.LogInternalError(`Visiting null`, messages_1.DielInternalErrorType.ArgNull);
        }
    }
    function _visitRelationReference(ref) {
        switch (ref.relationReferenceType) {
            case dielAstTypes_1.RelationReferenceType.Direct: {
                const r = ref;
                r.relationName = _changeString(r.relationName);
                return;
            }
            case dielAstTypes_1.RelationReferenceType.Subquery: {
                const r = ref;
                if (r.subquery)
                    _visitSelection(r.subquery);
                return;
            }
            default:
                messages_1.LogInternalError(``);
                return;
        }
    }
    function _visitJoinAst(j) {
        // no nested templates allowed
        if (j.templateSpec) {
            messages_1.ReportDielUserError(`No nested templates allowed`);
        }
        _visitRelationReference(j.relation);
        if (j.predicate)
            _visitExprAst(j.predicate);
    }
    function _visitCompositeSelectionUnit(ast) {
        ast.relation.columnSelections.map(c => _visitColumnSelection(c));
        if (ast.relation.baseRelation)
            _visitRelationReference(ast.relation.baseRelation);
        if (ast.relation.joinClauses)
            ast.relation.joinClauses.map(j => _visitJoinAst(j));
        if (ast.relation.whereClause)
            _visitExprAst(ast.relation.whereClause);
        if (ast.relation.groupByClause) {
            ast.relation.groupByClause.selections.map(c => _visitExprAst(c));
            if (ast.relation.groupByClause.predicate) {
                _visitExprAst(ast.relation.groupByClause.predicate);
            }
        }
        if (ast.relation.orderByClause)
            ast.relation.orderByClause.map(c => _visitOrderByAst(c));
        if (ast.relation.limitClause)
            _visitExprAst(ast.relation.limitClause);
    }
    function _visitSelection(subAst) {
        if (subAst) {
            subAst.compositeSelections.map(s => _visitCompositeSelectionUnit(s));
        }
        else {
            messages_1.LogInternalWarning(`Attempted to call __visitSelection on null`);
        }
    }
}
exports.TryToApplyTemplate = TryToApplyTemplate;
//# sourceMappingURL=applyTemplate.js.map