"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DielAstGetters_1 = require("../DielAstGetters");
const messages_1 = require("../../util/messages");
const dielAstTypes_1 = require("../../parser/dielAstTypes");
const passesHelper_1 = require("./passesHelper");
const dielUtils_1 = require("../../util/dielUtils");
function ArrangeInTopologicalOrder(ast) {
    if (!ast.depTree)
        return messages_1.LogInternalError(`the dependency should be defined already!`);
    const orderedNames = passesHelper_1.getTopologicalOrder(ast.depTree);
    // TMP
    const newRelations = [];
    orderedNames.map(n => {
        const def = DielAstGetters_1.GetRelationDef(ast, n);
        if (def)
            newRelations.push(def);
    });
    function checkDiff() {
        const topo = new Set(newRelations.map(n => n.rName));
        const relation = new Set(ast.relations.map(r => r.rName));
        const diff = dielUtils_1.SetSymmetricDifference(topo, relation);
        if (diff.size > 0)
            messages_1.LogInternalError(`Topological sort length did not match, ${diff}`);
    }
    // there are cases when relations that are not read by anyone is just not inserted, i.e. no deps etc.
    if (newRelations.length > ast.relations.length) {
        checkDiff;
    }
    else if (newRelations.length === ast.relations.length) {
        ast.relations = newRelations;
        return;
    }
    else {
        const diff = dielUtils_1.SetSymmetricDifference(new Set(newRelations.map(n => n.rName)), new Set(ast.relations.map(r => r.rName)));
        diff.forEach(d => {
            const def = DielAstGetters_1.GetRelationDef(ast, d);
            if (def)
                newRelations.push(def);
        });
        checkDiff;
        ast.relations = newRelations;
    }
}
exports.ArrangeInTopologicalOrder = ArrangeInTopologicalOrder;
/**
 * return the set of the relations that depent on the table passed in
 * TODO add depndsOn?: true and do another pass that uses transitive closure to figure out all dependnecies
 * @param depTree
 * @param rName
 * @param depndsOn the boolean is defaulted to true, if it's false, it's the other direction.
 */
function DeriveDependentRelations(depTree, rName) {
    const allDependencies = new Set();
    oneStep(rName, allDependencies);
    // recursively checks for dependencies
    function oneStep(rName, newDependencies) {
        // search through dependency
        let oldSet = new Set(newDependencies);
        for (let [key, value] of depTree) {
            const found = value.dependsOn.filter(d => d === rName);
            if (found.length > 0) {
                newDependencies.add(key);
            }
        }
        // set difference
        const diff = new Set([...newDependencies].filter(x => !oldSet.has(x)));
        if (diff.size > 0) {
            // need to run this on more dependencies
            diff.forEach((v) => {
                oneStep(v, newDependencies);
            });
        }
        return newDependencies;
    }
    return allDependencies;
}
exports.DeriveDependentRelations = DeriveDependentRelations;
function DeriveDepTreeFromSqlRelations(views, dynamic) {
    const depTree = new Map();
    const isDynamic = (rName) => {
        return dynamic.has(rName);
    };
    views.map(v => AddSingleDependency(depTree, v.selection, v.rName, isDynamic));
    return depTree;
}
exports.DeriveDepTreeFromSqlRelations = DeriveDepTreeFromSqlRelations;
/**
 * Get the set of relations that the view depends on
 * needs to iterate until we hit the original tables
 * @param viewName
 * @param depTree
 */
function DeriveOriginalRelationsAViewDependsOn(depTree, viewName) {
    let dep = depTree.get(viewName);
    let tables = new Set();
    if (dep && dep.dependsOn.length > 0) {
        // breadth first
        let toVisit = dep.dependsOn.slice();
        let visited = [viewName];
        let next;
        while (toVisit.length > 0) {
            next = toVisit.shift();
            if (next) {
                const dep = depTree.get(next);
                if (!dep) {
                    return messages_1.LogInternalError(`Dependency ${dep} not found`);
                }
                if (dep.isDynamic) {
                    tables.add(next);
                    visited.push(next);
                    if (dep.dependsOn.length !== 0) {
                        messages_1.LogInternalError(`dynamic tables should have zero dependencies!`);
                    }
                    continue;
                }
                else {
                    let children = dep.dependsOn;
                    children.forEach(child => {
                        if (toVisit.indexOf(next) === -1 && visited.indexOf(next) === -1) {
                            toVisit.push(child);
                        }
                    });
                    visited.push(next);
                }
            }
        }
    }
    return tables;
}
exports.DeriveOriginalRelationsAViewDependsOn = DeriveOriginalRelationsAViewDependsOn;
// ------------- BEGIN SETTER --------------------
function AddDepTree(ast) {
    const isDynamic = (rName) => {
        const found = ast.relations.find(r => r.rName === rName);
        return (found && found.relationType === "EventTable" /* EventTable */) ? true : false;
    };
    DielAstGetters_1.GetAllDerivedViews(ast).map(v => AddSingleDependency(ast.depTree, v.selection.compositeSelections, v.rName, isDynamic));
    return ast.depTree;
}
exports.AddDepTree = AddDepTree;
function AddSingleDependencyByDerivedRelation(ast, view) {
    const isDynamic = (rName) => {
        return DielAstGetters_1.IsRelationEvent(ast, rName);
    };
    AddSingleDependency(ast.depTree, view.selection.compositeSelections, view.rName, isDynamic);
}
exports.AddSingleDependencyByDerivedRelation = AddSingleDependencyByDerivedRelation;
// CompositeSelection |CompositeSelectionFinal
// the relation is dynamic if it's an event view
function AddSingleDependency(depTree, selection, rName, isDynamic) {
    // first add dependency one way, then the other way
    const dependsOn = addDependencyOneWay(depTree, selection, rName, isDynamic);
    addDependencyOtherWay(depTree, dependsOn, rName, isDynamic);
}
exports.AddSingleDependency = AddSingleDependency;
// CompositeSelection |CompositeSelectionFinal
// incremental dep tree building
function addDependencyOneWay(depTree, selection, rName, isDynamic) {
    let dependsOn = [];
    selection.map(c => {
        const deps = getSelectionUnitDep(c.relation);
        dependsOn = deps.concat(dependsOn);
    });
    depTree.set(rName, {
        relationName: rName,
        isDynamic: isDynamic(rName),
        dependsOn,
        isDependedBy: []
    });
    return dependsOn;
}
// recursive!
// SelectionUnit | SelectionUnitFinal
function getSelectionUnitDep(s) {
    if (!s.baseRelation) {
        return [];
    }
    const depsRaw = getRelationReferenceDep(s.baseRelation);
    let deps = depsRaw ? depsRaw : [];
    // the predicates on joins might have dependencies too... #FIXMELATER
    if (s.joinClauses)
        s.joinClauses.map(j => {
            const joinDepsRaw = getRelationReferenceDep(j.relation);
            if (joinDepsRaw)
                deps = deps.concat(joinDepsRaw);
        });
    if (s.whereClause) {
        getExprDep(deps, s.whereClause);
    }
    return deps;
}
exports.getSelectionUnitDep = getSelectionUnitDep;
function getRelationReferenceDep(ref) {
    switch (ref.relationReferenceType) {
        case dielAstTypes_1.RelationReferenceType.Direct: {
            const r = ref;
            return [r.relationName];
        }
        case dielAstTypes_1.RelationReferenceType.Subquery: {
            const r = ref;
            let acc = [];
            r.subquery.compositeSelections.map(c => {
                acc = acc.concat(getSelectionUnitDep(c.relation));
            });
            return acc;
        }
    }
}
exports.getRelationReferenceDep = getRelationReferenceDep;
function getExprDep(depAcc, e) {
    if (!e) {
        debugger;
    }
    switch (e.exprType) {
        case dielAstTypes_1.ExprType.Relation:
            const relationExpr = e;
            relationExpr.selection.compositeSelections.map(newE => {
                depAcc.push(...getSelectionUnitDep(newE.relation));
            });
            break;
        case dielAstTypes_1.ExprType.Func:
            const whereFuncExpr = e;
            whereFuncExpr.args.map(newE => {
                getExprDep(depAcc, newE);
            });
            break;
        case dielAstTypes_1.ExprType.Parenthesis:
            getExprDep(depAcc, e.content);
            break;
        default:
            return;
        // do nothing for now
    }
}
function addDependencyOtherWay(depTree, dependsOn, viewName, isDynamic) {
    dependsOn.map(dO => {
        if (dO) {
            // avoid the case when its null
            // it's possible that these don't exist, if they are the leaves
            const dep = depTree.get(dO);
            if (!dep) {
                depTree.set(dO, {
                    relationName: dO,
                    dependsOn: [],
                    isDependedBy: [viewName],
                    isDynamic: isDynamic(dO),
                });
            }
            else {
                dep.isDependedBy.push(viewName);
            }
        }
    });
}
//# sourceMappingURL=dependency.js.map