"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const dielAstTypes_1 = require("../../parser/dielAstTypes");
const codeGenSql_1 = require("../../compiler/codegen/codeGenSql");
const dielAstTypes_2 = require("../../parser/dielAstTypes");
const antlr4ts_1 = require("antlr4ts");
const lexer = require("../../parser/grammar/DIELLexer");
const parser = require("../../parser/grammar/DIELParser");
const generateAst_1 = require("../../parser/generateAst");
const DielAstGetters_1 = require("../DielAstGetters");
function generateViewConstraintCheckQuery(query) {
    let ast = checkValidView(query);
    if (ast) {
        // valid view
        return checkViewConstraint(ast);
    }
    return null;
}
exports.generateViewConstraintCheckQuery = generateViewConstraintCheckQuery;
function viewConstraintCheck(ast) {
    return checkViewConstraint(ast);
}
exports.viewConstraintCheck = viewConstraintCheck;
// @LUCIE: this is already defined as a helper function in the compiler.ts file
// Check if this is a valid view query. Return ast if it is, or null.
function checkValidView(query) {
    const inputStream = new antlr4ts_1.ANTLRInputStream(query);
    const l = new lexer.DIELLexer(inputStream);
    const tokenStream = new antlr4ts_1.CommonTokenStream(l);
    const p = new parser.DIELParser(tokenStream);
    const tree = p.queries();
    let visitor = new generateAst_1.default();
    let ast = visitor.visitQueries(tree);
    if (ast.relations.length > 0) {
        return ast;
    }
    return null;
}
// Precondition: query is a valid view statement
// supports only a single relation in view
function checkViewConstraint(ast) {
    // var i, j;
    const ret = new Map();
    // Handling multiple view statements
    const views = DielAstGetters_1.GetAllDerivedViews(ast);
    for (let view of views) {
        let viewConstraint = view.constraints;
        let queries = [];
        let selClause;
        // Only when there is a constraint on view
        if (viewConstraint) {
            let compositeSelections = view.selection.compositeSelections;
            // 1. handle null constraint
            selClause = getSelectClauseAST(compositeSelections);
            let nullQueries = getNullQuery(viewConstraint, selClause);
            queries = queries.concat(nullQueries);
            // 2. handle unique constraint
            selClause = getSelectClauseAST(compositeSelections);
            let uniqueQueries = getUniqueQuery(viewConstraint, selClause);
            queries = queries.concat(uniqueQueries);
            // 3. handle check constraint
            selClause = getSelectClauseAST(compositeSelections);
            let checkQueries = getCheckQuery(viewConstraint, selClause);
            queries = queries.concat(checkQueries);
        }
        ret.set(view.rName, queries);
    }
    return ret;
}
exports.checkViewConstraint = checkViewConstraint;
/**
 * Takes in a selection clause ast and copy/re-format so that
 * it can go inside where clause.
 * e.g) where (select a1, a2 from t1 where a1 < 10)
 */
function getSelectClauseAST(fromSel) {
    const expr = {
        dataType: undefined,
        exprType: dielAstTypes_2.ExprType.Star
    };
    const columnSel = {
        expr
    };
    const baseRelation = {
        subquery: {
            compositeSelections: fromSel
        }
    };
    const selUnit = {
        isDistinct: undefined,
        columnSelections: [columnSel],
        baseRelation: baseRelation,
        derivedColumnSelections: [],
        joinClauses: [],
        whereClause: undefined,
        groupByClause: undefined,
        orderByClause: undefined,
        limitClause: undefined
    };
    return selUnit;
}
function getCheckQuery(viewConstraint, selUnit) {
    let exprAsts = viewConstraint.exprChecks;
    let ret = [];
    let whichConstraint;
    if (exprAsts && exprAsts.length > 0) {
        let i, exprAst, whereClause;
        // iterate over check constraints
        for (i = 0; i < exprAsts.length; i++) {
            // ast for where clause
            exprAst = exprAsts[i];
            // where in the query it was broken
            whichConstraint = "CHECK " + codeGenSql_1.GetSqlStringFromExpr(exprAst);
            whereClause = {
                exprType: dielAstTypes_2.ExprType.Func,
                dataType: dielAstTypes_1.DielDataType.Boolean,
                functionType: dielAstTypes_2.FunctionType.Custom,
                functionReference: "NOT",
                args: [exprAst.content] // strip out parenthesis
            };
            // ast for the whole clause
            selUnit.whereClause = whereClause;
            let str = codeGenSql_1.SqlStrFromSelectionUnit(selUnit);
            ret.push([str, whichConstraint]);
        }
    }
    return ret;
}
function getNullQuery(viewConstraint, selUnit) {
    // console.log(JSON.stringify(view_constraint, null, 2));
    let ret = [];
    if (viewConstraint.notNull && viewConstraint.notNull.length > 0) {
        let notNullColumns = viewConstraint.notNull;
        for (let i = 0; i < notNullColumns.length; i++) {
            // formating the AST for whereclause
            let whereClause = {
                exprType: dielAstTypes_2.ExprType.Func,
                dataType: dielAstTypes_1.DielDataType.Boolean,
                functionType: dielAstTypes_2.FunctionType.BuiltIn,
                functionReference: dielAstTypes_2.BuiltInFunc.ValueIsNull,
                args: []
            };
            // format argument AST
            let cname = notNullColumns[i];
            let whichConstraint = cname + " NOT NULL";
            const whereClauseArg = {
                exprType: dielAstTypes_2.ExprType.Column,
                columnName: cname
            };
            whereClause.args.push(whereClauseArg);
            // formatting the rest of the selection unit for Not NULl
            selUnit.whereClause = whereClause;
            // Generate proper query from AST
            let str = codeGenSql_1.SqlStrFromSelectionUnit(selUnit);
            ret.push([str, whichConstraint]);
        }
    }
    return ret;
}
function getUniqueQuery(viewConstraints, selUnit) {
    let ret = [];
    let uniques = viewConstraints.uniques;
    // check if unique constraint exists
    if (uniques && uniques.length > 0) {
        let str, i, groupbyArgs, groupByClause, predicateSel;
        for (i = 0; i < uniques.length; i++) {
            groupbyArgs = uniques[i];
            // which constraint it was broken
            let whichConstraint = `UNIQUE (${groupbyArgs})`;
            let groupbySelections = [];
            // format groupby AST
            // which coloum to group by
            groupbyArgs.map(function (colName) {
                let expr = {
                    exprType: dielAstTypes_2.ExprType.Column,
                    dataType: undefined,
                    hasStar: false,
                    columnName: colName
                };
                groupbySelections.push(expr);
            });
            // format having clause AST
            predicateSel = {
                exprType: dielAstTypes_2.ExprType.Func,
                functionType: dielAstTypes_2.FunctionType.Logic,
                functionReference: ">",
                dataType: dielAstTypes_1.DielDataType.Boolean,
                args: [
                    {
                        exprType: dielAstTypes_2.ExprType.Func,
                        dataType: dielAstTypes_1.DielDataType.Number,
                        functionType: dielAstTypes_2.FunctionType.Custom,
                        functionReference: "COUNT",
                        args: [{
                                exprType: dielAstTypes_2.ExprType.Column,
                                hasStar: true
                            }]
                    },
                    {
                        exprType: dielAstTypes_2.ExprType.Val,
                        dataType: dielAstTypes_1.DielDataType.Number,
                        value: 1
                    }
                ]
            };
            // format the whole groupby clause AST
            groupByClause = {
                selections: groupbySelections,
                predicate: predicateSel
            };
            selUnit.groupByClause = groupByClause;
            // change selUnit select clause
            let selectColumns = [];
            groupbySelections.map(expr => {
                selectColumns.push({ expr });
            });
            const cS = {
                expr: {
                    exprType: dielAstTypes_2.ExprType.Func,
                    dataType: dielAstTypes_1.DielDataType.Number,
                    functionType: dielAstTypes_2.FunctionType.Custom,
                    functionReference: "COUNT",
                    args: [{
                            exprType: dielAstTypes_2.ExprType.Star
                        }]
                }
            };
            selectColumns.push(cS);
            selUnit.columnSelections = selectColumns;
            // Generate proper query from AST
            str = codeGenSql_1.SqlStrFromSelectionUnit(selUnit);
            ret.push([str, whichConstraint]);
        }
    }
    return ret;
}
//# sourceMappingURL=generateViewConstraints.js.map