"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const messages_1 = require("../../util/messages");
const dielAstTypes_1 = require("../../parser/dielAstTypes");
const DielAstGetters_1 = require("../DielAstGetters");
const DielAstVisitors_1 = require("../DielAstVisitors");
function InferType(ast) {
    DielAstVisitors_1.WalkThroughSelectionUnits(ast, ((selection) => {
        _inferTypeForSelectionUnit(ast, selection);
    }));
}
exports.InferType = InferType;
function InferTypeForDerivedRelation(ast, view) {
    view.selection.compositeSelections.map(s => {
        _inferTypeForSelectionUnit(ast, s.relation);
    });
}
exports.InferTypeForDerivedRelation = InferTypeForDerivedRelation;
function _inferTypeForSelectionUnit(ast, selection) {
    if (!selection.derivedColumnSelections) {
        messages_1.LogInternalError(`Normalization pass of column selection should be defined before infer type is called! for derivedColumnSelections `);
    }
    selection.derivedColumnSelections.map(cs => {
        if (!cs.expr.dataType) {
            cs.expr.dataType = getTypeForExpr(ast, cs.expr, selection);
        }
    });
}
function getUdfType(ast, sUnit, funName, expr) {
    const normalizedName = funName.toLocaleUpperCase();
    switch (normalizedName) {
        case dielAstTypes_1.BuiltInFunc.Coalesce:
        case dielAstTypes_1.BuiltInFunc.IfThisThen:
            // 0 is the case clause, which is a boolean
            const firstExpr = expr.args[1];
            if (!firstExpr) {
                messages_1.LogInternalError(`If else then should contain the if clause, but is missing: ${JSON.stringify(expr, null, 2)}`);
            }
            return getTypeForExpr(ast, firstExpr, sUnit);
        default:
            const r = ast.udfTypes.find(u => u.udf === normalizedName);
            if (!r) {
                return messages_1.ReportDielUserError(`Type of ${funName} not defined. Original query is ${JSON.stringify(sUnit, null, 2)}.`);
            }
            return r.type;
    }
}
// recursive!
function getTypeForExpr(ast, expr, sUnit) {
    switch (expr.exprType) {
        case dielAstTypes_1.ExprType.Func:
            const funExpr = expr;
            return getUdfType(ast, sUnit, funExpr.functionReference, funExpr);
        case dielAstTypes_1.ExprType.Column:
            const columnExpr = expr;
            const cn = columnExpr.columnName;
            // case 1: check for keywords
            const special = dielAstTypes_1.BuiltInColumnTyppes.filter(sc => sc.column === cn)[0];
            if (special) {
                return special.type;
            }
            // make sure that the source is specified
            if (!columnExpr.relationName) {
                return messages_1.LogInternalError(`The normalization pass screwed up and did not specify the source relation for ${JSON.stringify(columnExpr)}`);
            }
            // case 2: see if its from the baseRelation
            if (sUnit.baseRelation) {
                if (columnExpr.relationName === sUnit.baseRelation.alias) {
                    return getColumnTypeFromReference(ast, columnExpr.columnName, sUnit.baseRelation);
                }
                // case 3: see if its from the joins
                if (sUnit.joinClauses) {
                    for (let i = 0; i < sUnit.joinClauses.length; i++) {
                        const j = sUnit.joinClauses[i];
                        if (columnExpr.relationName === j.relation.alias) {
                            return getColumnTypeFromReference(ast, columnExpr.columnName, j.relation);
                        }
                    }
                }
            }
            // FIXME: should pass some metadata for debugging
            debugger;
            messages_1.ReportDielUserError(`type not found for column`);
        case dielAstTypes_1.ExprType.Parenthesis:
            const parenExpr = expr;
            // unpack
            return getTypeForExpr(ast, parenExpr.content, sUnit);
        case dielAstTypes_1.ExprType.Val:
            const valExpr = expr;
            return valExpr.dataType;
        case dielAstTypes_1.ExprType.Star:
            return messages_1.LogInternalError(`Should only invoke or normalized selections`);
        default:
            return messages_1.LogInternalError(`Should have handled all cases`, messages_1.DielInternalErrorType.UnionTypeNotAllHandled);
    }
}
function getColumnTypeFromCompositionSelection(s, columnName) {
    const selections = s[0].relation.derivedColumnSelections;
    if (!selections)
        return messages_1.LogInternalError(`Should have done the normalization pass before this for column ${columnName} in composite selction:\n ${JSON.stringify(s, null, 2)}!\n`);
    for (let i = 0; i < selections.length; i++) {
        const s = selections[i];
        if (columnName === s.alias)
            return s.expr.dataType;
    }
    messages_1.LogInternalWarning(`Type not found for ${columnName}`);
    return null;
}
// the two types are interchangeable for access purposes.
// exported for tests!
function getColumnTypeFromRelation(relation, columnName) {
    // case 1: derived
    if (DielAstGetters_1.IsRelationTypeDerived(relation.relationType)) {
        return getColumnTypeFromCompositionSelection(relation.selection.compositeSelections, columnName);
    }
    // case 2: original
    const column = relation.columns.filter(r => r.cName === columnName);
    if (column.length > 0) {
        return column[0].dataType;
    }
    else {
        messages_1.LogInternalWarning(`Type not found for ${columnName}`);
        return null;
    }
}
exports.getColumnTypeFromRelation = getColumnTypeFromRelation;
function getColumnTypeFromReference(ast, columnName, ref) {
    switch (ref.relationReferenceType) {
        case dielAstTypes_1.RelationReferenceType.Direct: {
            const r = ref;
            // we need the original name, not alias!
            const rDef = DielAstGetters_1.GetRelationDef(ast, r.relationName);
            return getColumnTypeFromRelation(rDef, columnName);
        }
        case dielAstTypes_1.RelationReferenceType.Subquery: {
            const rDef = ref.subquery.compositeSelections;
            return getColumnTypeFromCompositionSelection(rDef, columnName);
        }
        default:
            return messages_1.LogInternalError(``);
    }
}
//# sourceMappingURL=inferType.js.map