"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const passesHelper_1 = require("./passesHelper");
const dependency_1 = require("./dependency");
const distributeQueries_1 = require("./distributeQueries");
const messages_1 = require("../../util/messages");
const dielAstTypes_1 = require("../../parser/dielAstTypes");
/**
 * For now we wil just set the changes on all original tables
 * @param ast
 */
function TransformAstForMaterialization(ast) {
    const dynamic = new Set(ast.relations.filter(r => r.isDynamic).map(r => r.rName));
    const views = ast.relations.filter(r => r.relationType === "View" /* View */);
    const deps = dependency_1.DeriveDepTreeFromSqlRelations(views, dynamic);
    // get topo order
    const topoOrder = passesHelper_1.getTopologicalOrder(deps);
    function getRelationDef(rName) {
        return ast.relations.find(v => v.rName === rName);
    }
    const toMaterialize = getRelationsToMateralize(deps, getRelationDef);
    let originalTables;
    // Materialize by topological order
    topoOrder.forEach(relation => {
        if (toMaterialize.indexOf(relation) !== -1) {
            const view = getRelationDef(relation);
            if (!view) {
                messages_1.LogInternalError(`${relation} was not found!`);
            }
            else {
                originalTables = dependency_1.DeriveOriginalRelationsAViewDependsOn(deps, view.rName);
                if (originalTables.size === 0) {
                    // this means that this is a static view, in that it just need to be populated once...
                    // we then just need to change its reference to a table
                    // set in place
                    view.relationType = "Table" /* Table */;
                }
                else {
                    materializeAView(view, ast, originalTables);
                }
            }
        }
    });
}
exports.TransformAstForMaterialization = TransformAstForMaterialization;
/**
 * Change the derived view ast into program ast in place
 */
function materializeAView(view, ast, originalTables) {
    const columns = distributeQueries_1.GetColumnsFromSelection(view.selection);
    if (!columns) {
        messages_1.LogInternalError(`Columsn for ${view.selection} undefined`);
        return;
    }
    let table = {
        relationType: "Table" /* Table */,
        rName: view.rName,
        columns
    };
    // 2. make a program ast
    // 2-1. create insert,delete ast
    let deleteCommand = makeDeleteCommand(view.rName);
    let insertCommand = makeInsertCommand(view);
    // 3. push into programs. (this is supposed to preserve topo order)
    originalTables.forEach(rName => {
        // if the tname already exists in the map, append the program
        const existingTrigger = ast.triggers.find(t => t.afterRelationName === rName);
        if (existingTrigger) {
            existingTrigger.commands.push(deleteCommand, insertCommand);
        }
        else {
            ast.triggers.push({
                tName: `${rName}Trigger`,
                afterRelationName: rName,
                commands: [deleteCommand, insertCommand]
            });
        }
    });
    // 4. build the final ast. The order of relations sometimes changes
    // since table -> view -> output order.
    let relationIndex = ast.relations.indexOf(view);
    ast.relations[relationIndex] = table;
}
/**
 * Translate view constraints to table constraints.
 * @param view
 * @param table
 */
function translateConstraints(view, table) {
    if (view.constraints) {
        // 1. translate column constraints
        table.columns.forEach(c => {
            // 1-1. Handle NOT NULL constraint
            if (view.constraints.notNull.indexOf(c.cName) !== -1) {
                c.constraints.notNull = true;
            }
            // 1-2. Handle UNIQUE column constraint
            view.constraints.uniques.forEach(array => {
                if (array.length === 1 && array[0] === c.cName) {
                    c.constraints.unique = true;
                }
            });
            // 1-3. No need to translate check constraints
            // they are directly copied in step 2, at the end.
        });
        // 2. copy relation constraints
        table.constraints = view.constraints;
    }
    else {
        table.constraints = {
            relationNotNull: false,
            relationHasOneRow: false,
            primaryKey: [],
            notNull: [],
            uniques: [],
            exprChecks: [],
            foreignKeys: [],
        };
    }
}
/**
 * Create AST for DeleteClause
 * e.g) delete from v2;
 * @param view
 */
function makeDeleteCommand(viewName) {
    let deleteClause;
    deleteClause = {
        astType: dielAstTypes_1.AstType.Delete,
        relationName: viewName,
        predicate: undefined
    };
    return deleteClause;
}
/**
 * Create AST for InsertClause
 * e.g) insert into v2 select a + 1 as aPrime from v1;
 * @param view
 */
function makeInsertCommand(view) {
    let insertClause;
    insertClause = {
        astType: dielAstTypes_1.AstType.Insert,
        relation: view.rName,
        columns: [],
        selection: {
            astType: dielAstTypes_1.AstType.RelationSelection,
            compositeSelections: view.selection
        }
    };
    return insertClause;
}
/**
 * take in dependency tree and a relation definition lookup function
 *          o1
 *         /
 * t1 -> v1 - o2
 * @param ast
 */
function getRelationsToMateralize(depTree, getRelationDef) {
    let toMAterialize = [];
    depTree.forEach((nodeDep, relationName) => {
        const rDef = getRelationDef(relationName);
        if (rDef && (rDef.relationType === "View" /* View */)) {
            // and if the view is dependent on by at least two views/outputs, mark it as to materialize
            if (nodeDep.isDependedBy.length > 1) {
                toMAterialize.push(relationName);
            }
        }
    });
    return toMAterialize;
}
//# sourceMappingURL=materialization.js.map