"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const messages_1 = require("../../util/messages");
const dielAstTypes_1 = require("../../parser/dielAstTypes");
const DielAstVisitors_1 = require("../DielAstVisitors");
const DielAstGetters_1 = require("../DielAstGetters");
/**
 * the pass removes the .* as well as filling in where the columns comes from if it's not specified
 * - visit by topological order
 * - supports subqueries, e.g., select k.* from (select * from t1) k;
 */
function NormalizeColumnSelection(ast) {
    DielAstVisitors_1.WalkThroughSelectionUnits(ast, normalizeColumnForSelectionUnit);
}
exports.NormalizeColumnSelection = NormalizeColumnSelection;
/**
 * Note that if the referenced relation is an event table, the relevant event timesteps will be included
 * @param ast
 * @param ref
 */
function columnsFromRelationReference(ast, ref) {
    switch (ref.relationReferenceType) {
        case dielAstTypes_1.RelationReferenceType.Direct: {
            const r = ref;
            const def = DielAstGetters_1.GetRelationDef(ast, r.relationName);
            return DielAstGetters_1.DeriveColumnsFromRelation(def);
        }
        case dielAstTypes_1.RelationReferenceType.Subquery: {
            const r = ref;
            const newSelUnit = r.subquery.compositeSelections[0].relation;
            normalizeColumnForSelectionUnit(newSelUnit, ast);
            return DielAstGetters_1.DeriveColumnsFromSelectionUnit(newSelUnit);
        }
        default:
            return messages_1.LogInternalError(`Case ${ref.relationReferenceType} not handled`, messages_1.DielInternalErrorType.UnionTypeNotAllHandled);
    }
}
function columnsFromLocalSelectionUnit(ast, s, refName) {
    if (refName === s.baseRelation.alias) {
        const baseResult = columnsFromRelationReference(ast, s.baseRelation);
        if (baseResult)
            return baseResult;
    }
    if (s.joinClauses) {
        for (let i = 0; i < s.joinClauses.length; i++) {
            const joinRef = s.joinClauses[i];
            if (refName === joinRef.relation.alias) {
                const joinResult = columnsFromRelationReference(ast, joinRef.relation);
                if (joinResult) {
                    return joinResult;
                }
            }
        }
    }
    return messages_1.ReportDielUserError(`Relation not defined`);
}
function starCase(ast, s, currentColumnExpr) {
    // both of the following options will need to append to the selections
    if (currentColumnExpr.relationName) {
        // case 1: find the columns of just the relations specified
        const populatedColumns = columnsFromLocalSelectionUnit(ast, s, currentColumnExpr.relationName)
            .map(newColumn => ({
            exprType: dielAstTypes_1.ExprType.Column,
            dataType: newColumn.dataType,
            hasStar: false,
            columnName: newColumn.cName,
            relationName: currentColumnExpr.relationName
        }));
        return populatedColumns;
    }
    else {
        // case 2: find all the relations
        const populatedColumnsFromBase = columnsFromRelationReference(ast, s.baseRelation)
            .map(newColumn => ({
            exprType: dielAstTypes_1.ExprType.Column,
            dataType: newColumn.dataType,
            columnName: newColumn.cName,
            relationName: s.baseRelation.alias
        }));
        let populatedColumnsFromJoins = [];
        s.joinClauses.map(j => {
            const relationName = j.relation.alias;
            const newColumns = columnsFromRelationReference(ast, j.relation).map(c => ({
                exprType: dielAstTypes_1.ExprType.Column,
                columnName: c.cName,
                relationName: relationName,
            }));
            populatedColumnsFromJoins.push(...newColumns);
        });
        return populatedColumnsFromBase.concat(populatedColumnsFromJoins);
    }
}
function NormalizeColumnForDerivedRelation(ast, view) {
    view.selection.compositeSelections.map(s => {
        normalizeColumnForSelectionUnit(s.relation, ast);
    });
}
exports.NormalizeColumnForDerivedRelation = NormalizeColumnForDerivedRelation;
function normalizeRelationReference(ref, ast, rName) {
    if (ref.relationReferenceType === dielAstTypes_1.RelationReferenceType.Subquery) {
        const r = ref;
        normalizeColumnForSelectionUnit(r.subquery.compositeSelections[0].relation, ast, rName);
    }
}
/**
 * modify in place : (
 * @param s
 * @param ast
 * @param rName optional, used for providing informative error messages.
 */
function normalizeColumnForSelectionUnit(s, ast, rName) {
    console.log("normalized column for ", rName);
    if (s.derivedColumnSelections) {
        // not an error
        return;
    }
    // 1. let's first deal with the subqueries
    if (s.baseRelation) {
        normalizeRelationReference(s.baseRelation, ast, rName);
        if (s.joinClauses) {
            s.joinClauses.map(j => {
                normalizeRelationReference(j.relation, ast, rName);
            });
        }
    }
    // 2. then the actual selections
    const derivedColumnSelections = [];
    for (let i = 0; i < s.columnSelections.length; i++) {
        const c = s.columnSelections[i];
        const exprs = normalizeExpr(ast, s, c.expr);
        if (exprs.length === 0) {
            messages_1.LogInternalError(`exprs should at least be 1`);
            return;
        }
        else if (exprs.length === 1) {
            derivedColumnSelections.push({
                expr: exprs[0],
                alias: c.alias
            });
        }
        else {
            // this must be the star case
            // if the stars were all named
            for (let j = 0; j < exprs.length; j++) {
                const expr = exprs[j];
                if (expr.exprType !== dielAstTypes_1.ExprType.Column) {
                    messages_1.LogInternalError(`Only columns should be expanded`);
                }
                const cE = expr;
                // if there is just one relation, we shouldn't...
                // we can potentially look into improving the experience by omitting the join?
                // const alias = cE.columnName;
                const alias = (s.joinClauses.length === 0) || (s.columnSelections.filter(s => s.expr.exprType === dielAstTypes_1.ExprType.Star).length === 1)
                    ? cE.columnName
                    : `${cE.relationName}_${cE.columnName}`;
                derivedColumnSelections.push({
                    expr,
                    alias
                });
            }
        }
    }
    s.derivedColumnSelections = derivedColumnSelections;
    return;
}
function normalizeExpr(ast, s, e) {
    switch (e.exprType) {
        case dielAstTypes_1.ExprType.Star: {
            return starCase(ast, s, e);
        }
        case dielAstTypes_1.ExprType.Column: {
            return [normalizeColumnExpr(ast, s, e)];
        }
        case dielAstTypes_1.ExprType.Func: {
            // we need to walk thru for the exprs in the arguments!
            // we need to copy this function ast and then modify it...
            const newColumn = JSON.parse(JSON.stringify(e));
            normalizeFuncExpr(ast, s, newColumn);
            return [newColumn];
        }
        case dielAstTypes_1.ExprType.Val:
            // no change, we can copy by reference
            return [e];
        case dielAstTypes_1.ExprType.Parenthesis: {
            // keep going
            return normalizeExpr(ast, s, e.content);
        }
        case dielAstTypes_1.ExprType.Relation:
        default:
            // might be slow? #FIXME
            // return [JSON.parse(JSON.stringify(c))];
            messages_1.LogInternalError(`Not all cases are handled: ${e.exprType}`);
    }
    return messages_1.LogInternalError(`Should have returned by now`);
}
/**
 * This is ALL recursive...
 * must be incoked on ColumnSelection with ExprFunAsts
 * @param ast
 * @param s
 * @param c: will be modifying in place
 */
function normalizeFuncExpr(ast, s, e) {
    let newArgs = [];
    for (let i = 0; i < e.args.length; i++) {
        const a = e.args[i];
        newArgs = newArgs.concat(normalizeExpr(ast, s, a));
    }
    e.args = newArgs;
    return e;
}
/**
 * resursive
 * does NOT set in place
 * @param ast
 * @param s
 * @param c: will be modifying in place
 */
function normalizeColumnExpr(ast, s, currentColumnExpr) {
    // if there is already a source specified, do NOT modify
    if (currentColumnExpr.relationName) {
        return JSON.parse(JSON.stringify(currentColumnExpr));
    }
    const capsName = currentColumnExpr.columnName.toUpperCase();
    // TODO: we should probably just normalize the table instead of catching corner cases here
    if (capsName in dielAstTypes_1.BuiltInColumn) {
        const dataType = dielAstTypes_1.BuiltInColumnDataTypes.get(capsName);
        function createBuiltInColumnsIfReferencesEvent(ref) {
            if (ref.relationReferenceType === dielAstTypes_1.RelationReferenceType.Direct) {
                const relationName = ref.relationName;
                if (DielAstGetters_1.IsRelationEvent(ast, relationName)) {
                    const newColumnExpr = {
                        exprType: dielAstTypes_1.ExprType.Column,
                        dataType,
                        columnName: currentColumnExpr.columnName,
                        relationName,
                    };
                    return newColumnExpr;
                }
            }
            return null;
        }
        const newColumn = createBuiltInColumnsIfReferencesEvent(s.baseRelation);
        if (newColumn)
            return newColumn;
        if (s.joinClauses) {
            for (let i = 0; i < s.joinClauses.length; i++) {
                const j = s.joinClauses[i];
                const newColumn = createBuiltInColumnsIfReferencesEvent(j.relation);
                if (newColumn)
                    return newColumn;
            }
        }
        // the name might actually used by a different relation that is not built in; continue
    }
    // otherwise we need to search thru the relations for custom column names
    // const existingType = GetRelationColumnType(deAliasedRelationname, cn);
    if (s.baseRelation) {
        const found = checkIfColumnInRelationReference(ast, s.baseRelation, currentColumnExpr.columnName);
        if (found)
            return {
                exprType: dielAstTypes_1.ExprType.Column,
                dataType: found.dataType,
                columnName: currentColumnExpr.columnName,
                relationName: s.baseRelation.alias,
            };
        if (s.joinClauses) {
            for (let i = 0; i < s.joinClauses.length; i++) {
                const jClause = s.joinClauses[i];
                const found = checkIfColumnInRelationReference(ast, jClause.relation, currentColumnExpr.columnName);
                if (found)
                    return {
                        exprType: dielAstTypes_1.ExprType.Column,
                        dataType: found.dataType,
                        columnName: currentColumnExpr.columnName,
                        relationName: jClause.relation.alias,
                    };
            }
        }
    }
    return messages_1.LogInternalError(`We cannot find ${currentColumnExpr.columnName} anywhere, maybe you have called on the wrong column name.`);
}
function checkIfColumnInRelationReference(ast, ref, columnName) {
    // first see if the relation reference is direct
    const columns = columnsFromRelationReference(ast, ref);
    if (columns)
        return columns.find(c => c.cName === columnName);
    return null;
}
//# sourceMappingURL=normalizeColumnSelection.js.map