"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const runtimeTypes_1 = require("./runtimeTypes");
const runtimeHelper_1 = require("./runtimeHelper");
const dielUtils_1 = require("../util/dielUtils");
const messages_1 = require("../util/messages");
const DielRemoteActionToEngineActionWorker = new Map([
    [runtimeTypes_1.DielRemoteAction.GetResultsByPromise, "exec"],
    [runtimeTypes_1.DielRemoteAction.Close, "close"],
    [runtimeTypes_1.DielRemoteAction.UpdateRelation, "exec"],
    [runtimeTypes_1.DielRemoteAction.ConnectToDb, "open"],
    [runtimeTypes_1.DielRemoteAction.DefineRelations, "exec"],
    [runtimeTypes_1.DielRemoteAction.ShipRelation, "exec"],
    [runtimeTypes_1.DielRemoteAction.CleanUpQueries, "cleanup"] // supported only for socket
]);
const DielRemoteActionToEngineActionSocket = new Map(DielRemoteActionToEngineActionWorker);
DielRemoteActionToEngineActionSocket.set(runtimeTypes_1.DielRemoteAction.DefineRelations, "run");
DielRemoteActionToEngineActionSocket.set(runtimeTypes_1.DielRemoteAction.UpdateRelation, "run");
// to unify worker and websocket
class ConnectionWrapper {
    constructor(connection, remoteType) {
        this.remoteType = remoteType;
        this.connection = connection;
        this.globalMsgId = 1;
        this.resolves = new Map();
        this.rejects = new Map();
    }
    send(id, msgToSend, isPromise) {
        messages_1.LogExecutionTrace(`Executing to DB ${msgToSend.sql}`); // REMOVE
        if (isPromise) {
            // do the promise thing here
            const msgId = this.globalMsgId;
            this.globalMsgId += 1;
            const self = this;
            return new Promise(function (resolve, reject) {
                // save callbacks for later
                self.resolves.set(msgId, resolve);
                self.rejects.set(msgId, reject);
                self.sendNoPromise(Object.assign({}, id, { msgId }), msgToSend);
            });
        }
        else {
            this.sendNoPromise(id, msgToSend);
            return null;
        }
    }
    sendNoPromise(id, msgToSend) {
        if (this.remoteType === runtimeTypes_1.DbType.Worker) {
            // FIXME: might need some adaptor logic here to make worker the same as socket
            const action = DielRemoteActionToEngineActionWorker.get(id.remoteAction);
            if (!action) {
                messages_1.LogInternalError(`Action must be defined, but not for ${id.remoteAction}`);
            }
            const worker = this.connection;
            // FIXME: JSON.stringify(id) ?
            const finalMsg = Object.assign({ id,
                action }, msgToSend);
            messages_1.LogExecutionTrace(`Posting to Worker`, finalMsg);
            worker.postMessage(finalMsg);
        }
        else {
            // FIXME: clear serialization logic
            const action = DielRemoteActionToEngineActionSocket.get(id.remoteAction);
            const newMessage = Object.assign({ id,
                action }, msgToSend);
            this.connection.send(JSON.stringify(newMessage));
        }
    }
    // FIXME: maybe don't need this since they have the same name
    setHandler(f) {
        const self = this;
        const newF = (event) => {
            messages_1.LogExecutionTrace(`DB executed message`, event.data);
            let msg;
            if (this.remoteType === runtimeTypes_1.DbType.Socket) {
                try {
                    msg = parseDielReply(event.data);
                    // special debugging case
                    if (msg.id === "test")
                        return;
                }
                catch (e) {
                    messages_1.LogInternalWarning(`Socket sent mal-formatted message: ${JSON.stringify(event.data, null, 2)}`);
                    return;
                }
            }
            else {
                if (event.data.id) {
                    // special debugging case
                    if (event.data.id === "test")
                        return;
                    if (event.data.id === "download") {
                        let blob = new Blob([event.data.buffer]);
                        dielUtils_1.downloadHelper(blob, "workerSession");
                        return;
                    }
                    msg = {
                        id: event.data.id,
                        results: runtimeHelper_1.ParseSqlJsWorkerResult(event.data.results),
                        err: event.data.err
                    };
                }
                else {
                    console.log(`%c\nWorker sent mal-formatted message: ${event.data}`, "color: red");
                    return;
                }
            }
            if (msg.id.msgId) {
                const promiseId = msg.id.msgId;
                if (msg.err) {
                    // error condition
                    const reject = self.rejects.get(promiseId);
                    if (reject) {
                        reject(msg.err);
                    }
                }
                else {
                    const resolve = self.resolves.get(promiseId);
                    if (resolve) {
                        resolve(msg);
                    }
                }
                // LogExecutionTrace("Promise resolved", 1, msg);
                // purge used callbacks
                self.resolves.delete(promiseId);
                self.rejects.delete(promiseId);
            }
            f(msg);
        };
        if (this.remoteType === runtimeTypes_1.DbType.Worker) {
            this.connection.onmessage = newF;
        }
        else {
            this.connection.onmessage = newF;
        }
    }
}
exports.ConnectionWrapper = ConnectionWrapper;
function parseDielReply(rawStr) {
    let msg = JSON.parse(rawStr);
    // msg.id = JSON.parse(msg.id);
    return msg;
}
//# sourceMappingURL=ConnectionWrapper.js.map