"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const runtimeTypes_1 = require("./runtimeTypes");
const DielRuntime_1 = require("./DielRuntime");
const messages_1 = require("../util/messages");
const DielPhysicalExecution_1 = require("../compiler/DielPhysicalExecution");
const dielUtils_1 = require("../util/dielUtils");
const ConnectionWrapper_1 = require("./ConnectionWrapper");
function connectToSocket(url) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise(function (resolve, reject) {
            let socket = new WebSocket(url);
            socket.onopen = function () {
                console.log("resolved socket");
                resolve(socket);
            };
            socket.onerror = function (err) {
                console.log("socket error");
                reject(err);
            };
        });
    });
}
class DbEngine {
    constructor(config, remoteId, relationShippingCallback) {
        // this.remoteType = remoteType;
        this.config = config;
        this.id = remoteId;
        this.totalMsgCount = 0;
        this.msgCountTimeStamp = Date.now();
        this.queueMap = new Map();
        this.relationShippingCallback = relationShippingCallback;
    }
    Close() {
        switch (this.config.dbType) {
            case runtimeTypes_1.DbType.Worker:
                const id = {
                    remoteAction: runtimeTypes_1.DielRemoteAction.Close,
                    // timestep not needed...
                    requestTimestep: -1
                };
                // the sql is also not needed
                this.connection.send(id, { sql: "" }, false);
            case runtimeTypes_1.DbType.Socket:
        }
    }
    setup() {
        return __awaiter(this, void 0, void 0, function* () {
            switch (this.config.dbType) {
                case runtimeTypes_1.DbType.Worker:
                    const configWorker = this.config;
                    let newConnection;
                    try {
                        newConnection = new Worker(configWorker.jsFile);
                    }
                    catch (e) {
                        return messages_1.ReportDielUserError(`Error loading worker: ${e}. The path used is ${configWorker.jsFile}.`);
                    }
                    this.connection = new ConnectionWrapper_1.ConnectionWrapper(newConnection, this.config.dbType);
                    // note that this must be set before the await is called, otherwise we get into a dealock!
                    // same code for socket for below as well
                    this.connection.setHandler(this.getHandleMsgForRemote());
                    const response = yield fetch(configWorker.dataFile);
                    const bufferRaw = yield response.arrayBuffer();
                    const buffer = new Uint8Array(bufferRaw);
                    // we should block because if it's not ack-ed the rest of the messages cannot be processed properly
                    return yield this.SendMsg({
                        remoteAction: runtimeTypes_1.DielRemoteAction.ConnectToDb,
                        requestTimestep: DielRuntime_1.INIT_TIMESTEP,
                        buffer,
                    }, true);
                case runtimeTypes_1.DbType.Socket:
                    const configSocket = this.config;
                    try {
                        const socket = yield connectToSocket(configSocket.connection);
                        this.connection = new ConnectionWrapper_1.ConnectionWrapper(socket, this.config.dbType);
                        this.connection.setHandler(this.getHandleMsgForRemote());
                        if (configSocket.message) {
                            return yield this.SendMsg({
                                remoteAction: runtimeTypes_1.DielRemoteAction.ConnectToDb,
                                requestTimestep: DielRuntime_1.INIT_TIMESTEP,
                                message: configSocket.message
                            }, true);
                        }
                        // FIXME
                        return messages_1.LogInternalError(`Should not be in this condition`);
                    }
                    catch (e) {
                        return messages_1.ReportDielUserError(`Failed to connect to socket at ${configSocket.connection}, with error: ${e}.`);
                    }
                case runtimeTypes_1.DbType.Local:
                    return messages_1.LogInternalError(`Should not use DbEngine wrapper for local`);
                default:
                    return messages_1.LogInternalError(`handle different connections`, messages_1.DielInternalErrorType.UnionTypeNotAllHandled);
            }
        });
    }
    nextQueue() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.currentQueueHead)
                this.queueMap.delete(this.currentQueueHead);
            this.currentQueueHead = null;
            // if there is more in the queue to address, deal with them
            if (this.queueMap.size > 0) {
                this.currentQueueHead = Math.min(...this.queueMap.keys());
                // also load the next queue's results in
                const currentQueueMap = this.queueMap.get(this.currentQueueHead);
                if (!currentQueueMap) {
                    messages_1.LogInternalError(`CurrentQueneMap not defined`);
                    return;
                }
                const promises = currentQueueMap.receivedValues.map((updateMsg) => __awaiter(this, void 0, void 0, function* () {
                    const id = {
                        remoteAction: updateMsg.remoteAction,
                        msgId: updateMsg.msgId,
                        requestTimestep: updateMsg.requestTimestep,
                    };
                    const msgToSend = this.extendMsgWithCustom({
                        sql: updateMsg.sql,
                    });
                    yield this.connection.send(id, msgToSend, true);
                }));
                yield Promise.all(promises);
                // note that this must be evaluated after the messages have actually finished executing
                // which means asynchrony!
                yield this.evaluateQueueOnUpdateHandler();
            }
            else {
                // debugger;
                messages_1.LogExecutionTrace(`Setting current queue head to null`);
            }
        });
    }
    // RECURSIVE
    evaluateQueueOnUpdateHandler() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.currentQueueHead) {
                // if there is no queue, skip
                if (this.queueMap.size === 0) {
                    return null;
                }
                const queueKeys = this.queueMap.keys();
                // this is the first time
                this.currentQueueHead = Math.min(...queueKeys);
            }
            const currentItem = this.queueMap.get(this.currentQueueHead);
            if (!currentItem) {
                return messages_1.LogInternalError(`Queue should contain current head ${this.currentQueueHead}, but it contains ${this.queueMap.keys()}!`);
            }
            messages_1.LogExecutionTrace(`Attempting to process queue at ${this.currentQueueHead}`, this.queueMap);
            // coarse grained
            const currentQueueHead = this.currentQueueHead;
            if (!currentQueueHead)
                return messages_1.LogInternalError(`Curretn quene head shoudl be defined!`);
            if (dielUtils_1.IsSetIdentical(currentItem.received, currentItem.deps)) {
                console.log(`  Success Processing queue at ${this.currentQueueHead}`, currentItem.relationsToShip);
                yield currentItem.relationsToShip.forEach((destinations, relationName) => __awaiter(this, void 0, void 0, function* () {
                    yield destinations.forEach((dbId) => __awaiter(this, void 0, void 0, function* () {
                        const shipMsg = {
                            remoteAction: runtimeTypes_1.DielRemoteAction.ShipRelation,
                            requestTimestep: currentQueueHead,
                            relationName,
                            dbId
                        };
                        yield this.SendMsg(shipMsg, true);
                    }));
                }));
                // let's make sure that this is sent...
                // debugger;
                this.nextQueue();
                return null;
            }
            else {
                console.log(`  Failure! Two sets are not the same`, currentItem.received, currentItem.deps);
                // need to keep on waiting
                return null;
            }
        });
    }
    extendMsgWithCustom(msg) {
        switch (this.config.dbType) {
            case runtimeTypes_1.DbType.Worker:
                return msg;
            case runtimeTypes_1.DbType.Socket:
                return Object.assign({ message: this.config.message }, msg);
            default:
                messages_1.LogInternalError(`Shouldn't be any other cases`);
        }
    }
    SendMsg(msg, isPromise = false) {
        messages_1.LogExecutionTrace(`Received message. Worker attempting to action ${msg.remoteAction} for request step ${msg.requestTimestep}`, msg);
        this.totalMsgCount++;
        this.sanityCheck();
        const id = {
            remoteAction: msg.remoteAction,
            requestTimestep: msg.requestTimestep,
        };
        switch (msg.remoteAction) {
            case runtimeTypes_1.DielRemoteAction.ConnectToDb: {
                const opMsg = msg;
                if (this.config.dbType === runtimeTypes_1.DbType.Worker) {
                    const buffer = opMsg.buffer;
                    const msgToSend = {
                        buffer
                    };
                    return this.connection.send(id, msgToSend, isPromise);
                }
                else {
                    const message = opMsg.message ? opMsg.message : "";
                    const msgToSend = {
                        message
                    };
                    return this.connection.send(id, msgToSend, isPromise);
                }
            }
            case runtimeTypes_1.DielRemoteAction.CleanUpQueries: {
                const cleanupMsg = msg;
                const msgToSend = this.extendMsgWithCustom({
                    sql: cleanupMsg.sql
                });
                return this.connection.send(id, msgToSend, isPromise);
            }
            case runtimeTypes_1.DielRemoteAction.UpdateRelation: {
                if (!this.physicalExeuctionRef)
                    return messages_1.LogInternalError(`should have reference to phsyical execution by now`);
                const updateMsg = msg;
                // push this on to the message queue
                if (this.queueMap.has(updateMsg.requestTimestep)) {
                    const requestTimestepQueue = this.queueMap.get(updateMsg.requestTimestep);
                    if (!requestTimestepQueue) {
                        return messages_1.LogInternalError(``);
                    }
                    requestTimestepQueue.received.add(updateMsg.relationName);
                }
                else {
                    const rToShip = this.physicalExeuctionRef.getRelationsToShipForDb(this.id, msg.requestTimestep);
                    if (!rToShip) {
                        return messages_1.LogInternalError(``);
                    }
                    this.queueMap.set(msg.requestTimestep, {
                        receivedValues: [],
                        received: new Set([updateMsg.relationName]),
                        relationsToShip: rToShip.relationsToShip,
                        deps: rToShip.deps
                    });
                    messages_1.LogExecutionTrace(`Setting queue with new request timestep ${msg.requestTimestep}`);
                }
                // then process
                // the following is brittle, but if we don't set it, the queue just keeps going
                if (!this.currentQueueHead)
                    this.currentQueueHead = msg.requestTimestep;
                if (msg.requestTimestep === this.currentQueueHead) {
                    messages_1.LogExecutionTrace(`Executing immediately as received, with request timestep: ${msg.requestTimestep}`);
                    // can actually execute execute
                    // figure out the dbname if it's there.
                    const msgToSend = this.extendMsgWithCustom({
                        sql: updateMsg.sql
                    });
                    return this.connection.send(id, msgToSend, isPromise);
                }
                else {
                    messages_1.LogExecutionTrace(`CANNOT execute immediately, pushing request timestep: ${msg.requestTimestep} to queue`);
                    // otherwise push on the queue
                    const requestTimestepQueue = this.queueMap.get(updateMsg.requestTimestep);
                    if (!requestTimestepQueue) {
                        return messages_1.LogInternalError(`requestTimestepQuee null`);
                    }
                    requestTimestepQueue.receivedValues.push(updateMsg);
                    return null;
                }
            }
            case runtimeTypes_1.DielRemoteAction.DefineRelations: {
                const defineMsg = msg;
                const msgToSend = this.extendMsgWithCustom({
                    sql: defineMsg.sql,
                });
                messages_1.LogSetup(`Running Query in Remote[${this.id}]:\n${defineMsg.sql}`);
                return this.connection.send(id, msgToSend, isPromise);
            }
            case runtimeTypes_1.DielRemoteAction.ShipRelation: {
                if (!this.physicalExeuctionRef)
                    return messages_1.LogInternalError(`should have reference to phsyical execution by now`);
                const shipMsg = msg;
                const newId = Object.assign({}, id, { relationName: shipMsg.relationName });
                // if (isPromise) {
                //   LogInternalError(`You cannot wait on ${DielRemoteAction.ShipRelation}`);
                // }
                // also need to check if it's being shipped to itself, in which case, we need to bubble up
                // this is assumed to be only for static relations... might be wrong
                if (shipMsg.dbId === this.id) {
                    // note that this is a good hook for materialization
                    // LogInfo(`Shipping to itself, this is a local evaluation`);
                    messages_1.LogInternalError(`why is this happening...`);
                    // in theory I think we can just invoke the connection directly.
                    // staticShip.map(t => {
                    //   const staticMsg: RemoteShipRelationMessage = {
                    //     remoteAction: DielRemoteAction.ShipRelation,
                    //     relationName: t.relation,
                    //     dbId: t.destination,
                    //     requestTimestep: msg.requestTimestep,
                    //   };
                    //   this.SendMsg(staticMsg);
                    // });
                    // return null;
                }
                if ((this.id !== DielPhysicalExecution_1.LocalDbId) && (shipMsg.dbId !== DielPhysicalExecution_1.LocalDbId)) {
                    // this case is not yet supported
                    messages_1.LogInternalError(`Shipping across remote engines from ${this.id} to ${shipMsg.dbId}`, messages_1.DielInternalErrorType.NotImplemented);
                    return null;
                }
                const msgToSend = this.extendMsgWithCustom({
                    sql: `select * from ${shipMsg.relationName};`,
                });
                return this.connection.send(newId, msgToSend, isPromise);
            }
            case runtimeTypes_1.DielRemoteAction.GetResultsByPromise: {
                const sql = msg.sql;
                const msgToSend = this.extendMsgWithCustom({
                    sql,
                });
                return this.connection.send(id, msgToSend, true);
            }
            default:
                return messages_1.LogInternalError(`DielRemoteAction ${msg.remoteAction} not handled`);
        }
    }
    /**
     * This function generates the handler for when the original messages (i.e., actions to the databases)
     *   have been completed.
     */
    getHandleMsgForRemote() {
        const handleMsg = (msg) => __awaiter(this, void 0, void 0, function* () {
            // if this is socket we need to unpack all, but if this is worker, then we just need to unpack id... so annoying
            switch (msg.id.remoteAction) {
                case runtimeTypes_1.DielRemoteAction.UpdateRelation: {
                    yield this.evaluateQueueOnUpdateHandler();
                    messages_1.LogExecutionTrace(`Finished UpdateRelation and queue evaluation`);
                    break;
                }
                case runtimeTypes_1.DielRemoteAction.GetResultsByPromise: {
                    console.log(`Db get results for ${this.id} done.`);
                    break;
                }
                case runtimeTypes_1.DielRemoteAction.ShipRelation: {
                    const view = msg.id.relationName;
                    if (!view || !msg.id.requestTimestep) {
                        messages_1.LogInternalError(`Both view and request_timestep should be defined for sharing views! However, I got ${JSON.stringify(msg.id, null, 2)}`);
                        return;
                    }
                    if (msg.results.length > 0) {
                        this.relationShippingCallback(view, msg.results, msg.id.requestTimestep);
                    }
                    break;
                }
                case runtimeTypes_1.DielRemoteAction.DefineRelations: {
                    console.log(`Relations defined successfully for ${this.id}`);
                    break;
                }
                case runtimeTypes_1.DielRemoteAction.CleanUpQueries: {
                    console.log(`Cleanup queries defined successfully for ${this.id}`);
                    break;
                }
                case runtimeTypes_1.DielRemoteAction.ConnectToDb: {
                    console.log(`Db opened for ${this.id}`);
                    break;
                }
                default:
                    messages_1.LogInternalError(`You should handle ${msg.id.remoteAction} as well`, messages_1.DielInternalErrorType.UnionTypeNotAllHandled);
            }
        });
        return handleMsg;
    }
    // hack to prevent loops in messaging logic
    sanityCheck() {
        // if we send 100 messages within 1 second, probably some loop going on...
        if (this.totalMsgCount > 100) {
            if (Date.now() - this.msgCountTimeStamp < 1000) {
                messages_1.LogInternalError(`Too many messages:`);
            }
            else {
                this.totalMsgCount = 0;
                this.msgCountTimeStamp = Date.now();
            }
        }
    }
    setPhysicalExecutionReference(physicalExeuctionRef) {
        console.log("setting physical exeuction reference", this.id);
        this.physicalExeuctionRef = physicalExeuctionRef;
    }
    downloadDb() {
        // only available on workers
        if (this.config.dbType === runtimeTypes_1.DbType.Worker) {
            this.connection.connection.postMessage({
                id: "download",
                action: "export",
            });
        }
        else {
            messages_1.LogInternalWarning(`You cannot download anything that's not Workers here`);
        }
    }
    /**
     * SqlitemasterQuery returns sql and name
     */
    getMetaData(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const promise = this.SendMsg({
                remoteAction: runtimeTypes_1.DielRemoteAction.GetResultsByPromise,
                requestTimestep: DielRuntime_1.INIT_TIMESTEP,
                sql: DielRuntime_1.SqliteMasterQuery
            }, true);
            const data = yield promise;
            return {
                id,
                data: data ? data.results : null
            };
        });
    }
}
exports.default = DbEngine;
//# sourceMappingURL=DbEngine.js.map